#!/usr/bin/env python3
"""
Network Protocol Dissector
Parameters:
    - IP: Interface IP to monitor
Additional parameters in comments:
    - protocol: Protocol to analyze (e.g., TCP, UDP, ICMP)
Requires: scapy package and admin/root privileges
"""
from scapy.all import *
import argparse
from collections import defaultdict
import json
from datetime import datetime

class ProtocolDissector:
    def __init__(self, interface_ip, protocol=None):
        self.interface_ip = interface_ip
        self.protocol = protocol
        self.stats = defaultdict(lambda: {
            'packet_count': 0,
            'byte_count': 0,
            'protocols': defaultdict(int),
            'ports': defaultdict(int),
            'packet_sizes': [],
            'timestamps': []
        })
        
    def analyze_tcp(self, packet):
        """Analyze TCP packet details"""
        tcp = packet[TCP]
        details = {
            'src_port': tcp.sport,
            'dst_port': tcp.dport,
            'seq': tcp.seq,
            'ack': tcp.ack,
            'flags': {
                'SYN': bool(tcp.flags & 0x02),
                'ACK': bool(tcp.flags & 0x10),
                'FIN': bool(tcp.flags & 0x01),
                'RST': bool(tcp.flags & 0x04),
                'PSH': bool(tcp.flags & 0x08)
            },
            'window': tcp.window
        }
        
        # Analyze TCP options
        options = {}
        for opt in tcp.options:
            if opt[0] == 'MSS':
                options['MSS'] = opt[1]
            elif opt[0] == 'WScale':
                options['WindowScale'] = opt[1]
            elif opt[0] == 'Timestamp':
                options['Timestamp'] = opt[1]
                
        if options:
            details['options'] = options
            
        return details
        
    def analyze_udp(self, packet):
        """Analyze UDP packet details"""
        udp = packet[UDP]
        return {
            'src_port': udp.sport,
            'dst_port': udp.dport,
            'length': udp.len
        }
        
    def analyze_icmp(self, packet):
        """Analyze ICMP packet details"""
        icmp = packet[ICMP]
        return {
            'type': icmp.type,
            'code': icmp.code,
            'id': icmp.id if hasattr(icmp, 'id') else None,
            'seq': icmp.seq if hasattr(icmp, 'seq') else None
        }
        
    def analyze_dns(self, packet):
        """Analyze DNS packet details"""
        dns = packet[DNS]
        details = {
            'id': dns.id,
            'qr': dns.qr,  # 0 for query, 1 for response
            'opcode': dns.opcode,
            'rcode': dns.rcode,
            'qdcount': dns.qdcount,
            'ancount': dns.ancount
        }
        
        # Add query details
        if dns.qd:
            details['query'] = {
                'name': dns.qd.qname.decode(),
                'type': dns.qd.qtype
            }
            
        # Add answer details
        if dns.an:
            details['answers'] = []
            for answer in dns.an:
                details['answers'].append({
                    'name': answer.rrname.decode(),
                    'type': answer.type,
                    'data': str(answer.rdata)
                })
                
        return details
        
    def packet_callback(self, packet):
        """Process each captured packet"""
        if not IP in packet:
            return
            
        # Basic packet info
        ip = packet[IP]
        proto = ip.proto
        length = len(packet)
        timestamp = time.time()
        
        # Update general statistics
        direction = 'outbound' if ip.src == self.interface_ip else 'inbound'
        stats = self.stats[direction]
        stats['packet_count'] += 1
        stats['byte_count'] += length
        stats['packet_sizes'].append(length)
        stats['timestamps'].append(timestamp)
        
        # Detailed protocol analysis
        details = {
            'timestamp': datetime.fromtimestamp(timestamp).isoformat(),
            'length': length,
            'ip': {
                'version': ip.version,
                'src': ip.src,
                'dst': ip.dst,
                'ttl': ip.ttl,
                'proto': ip.proto
            }
        }
        
        # Protocol-specific analysis
        if TCP in packet:
            details['tcp'] = self.analyze_tcp(packet)
            stats['protocols']['TCP'] += 1
            stats['ports'][f"TCP_{packet[TCP].dport}"] += 1
        elif UDP in packet:
            details['udp'] = self.analyze_udp(packet)
            stats['protocols']['UDP'] += 1
            stats['ports'][f"UDP_{packet[UDP].dport}"] += 1
            if DNS in packet:
                details['dns'] = self.analyze_dns(packet)
                stats['protocols']['DNS'] += 1
        elif ICMP in packet:
            details['icmp'] = self.analyze_icmp(packet)
            stats['protocols']['ICMP'] += 1
            
        # Print packet summary
        print(f"\rPackets captured: {sum(s['packet_count'] for s in self.stats.values())}", end="")
        
        # Detailed logging for specific protocols
        if self.protocol:
            if self.protocol.upper() in details:
                print(f"\nPacket Details: {json.dumps(details, indent=2)}")
    
    def print_statistics(self):
        """Print captured statistics"""
        print("\n\nProtocol Analysis Report")
        print("=" * 50)
        print(f"Interface IP: {self.interface_ip}")
        print(f"Timestamp: {datetime.now()}")
        print("=" * 50)
        
        for direction, stats in self.stats.items():
            print(f"\n{direction.upper()} Traffic:")
            print("-" * 40)
            print(f"Total Packets: {stats['packet_count']}")
            print(f"Total Bytes: {stats['byte_count']}")
            
            if stats['packet_sizes']:
                avg_size = sum(stats['packet_sizes']) / len(stats['packet_sizes'])
                print(f"Average Packet Size: {avg_size:.2f} bytes")
                
            print("\nProtocol Distribution:")
            for proto, count in stats['protocols'].items():
                percentage = (count / stats['packet_count']) * 100
                print(f"  {proto}: {count} ({percentage:.2f}%)")
                
            print("\nTop Ports:")
            sorted_ports = sorted(stats['ports'].items(),
                                key=lambda x: x[1], reverse=True)[:10]
            for port, count in sorted_ports:
                percentage = (count / stats['packet_count']) * 100
                print(f"  {port}: {count} ({percentage:.2f}%)")
    
    def start_capture(self):
        """Start packet capture and analysis"""
        try:
            print(f"Starting protocol analysis on {self.interface_ip}")
            if self.protocol:
                print(f"Focusing on {self.protocol.upper()} protocol")
            print("Press Ctrl+C to stop and show statistics...")
            
            # Create capture filter
            filter_str = f"host {self.interface_ip}"
            if self.protocol:
                filter_str += f" and {self.protocol.lower()}"
                
            # Start packet capture
            sniff(filter=filter_str, prn=self.packet_callback)
            
        except KeyboardInterrupt:
            self.print_statistics()
        except Exception as e:
            print(f"Error during capture: {str(e)}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Network Protocol Dissector")
    parser.add_argument("ip", help="Interface IP to monitor")
    parser.add_argument("--protocol", choices=['TCP', 'UDP', 'ICMP'],
                      help="Protocol to analyze in detail")
    args = parser.parse_args()
    
    dissector = ProtocolDissector(args.ip, args.protocol)
    dissector.start_capture()
